package org.openjdk.gcbench.runtime.cmpxchg;

import org.openjdk.jmh.annotations.*;
import sun.misc.Unsafe;

import java.lang.reflect.Field;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
@Threads(1)
@State(Scope.Benchmark)
public class WeakCasLoop {

    static final Unsafe U;
    static final long OFF_REF;

    static {
        try {
            Field unsafeField = Unsafe.class.getDeclaredField("theUnsafe");
            unsafeField.setAccessible(true);
            U = (Unsafe) unsafeField.get(null);
            OFF_REF  = U.objectFieldOffset(WeakCasLoop.class.getDeclaredField("ref"));
        } catch (Exception e) {
            throw new AssertionError(e);
        }
    }

    volatile Point ref;

    @Setup
    public void setup() {
        ref = new Point(0, 0);
    }

    @Benchmark
    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    public void strong() {
        Point ep, np;
        do {
            ep = ref;
            np = ep.next();
        } while (!U.compareAndSwapObject(this, OFF_REF, ep, np));
    }

    public static class Point {
        final int x;
        final int y;

        public Point(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public Point next() {
            return new Point((x + 1) & 255, (y + 2) & 255);
        }
    }

    /*
        i7 4790K, 4.0 Ghz, Linux x86_64, JDK 9 (Shenandoah, 2016-09-05)

            Benchmark         Mode  Cnt   Score   Error  Units

            # Shenandoah
            CasBarriers.test  avgt   25  16.418 ± 0.029  ns/op

            # G1
            CasBarriers.test  avgt   25  12.545 ± 0.041  ns/op

            # Parallel
            CasBarriers.test  avgt   25  12.526 ± 0.030  ns/op

        Analysis pending.
     */

}
